
typedef enum {									// IIC errors
	ie_ok,
	ie_start_high,
	ie_start_low,
	ie_stop_low,
	ie_stop_high
} T_IIC_ERROR;

template <										// --- Software IIC template
	const uint8_t scl,							// Clock line bitmask
	const uint8_t sda,							// Data line bitmask
	volatile unsigned char &pin,				// Input port
	volatile unsigned char &pout,				// Output port
	volatile unsigned char &pdir				// Port direction
> struct CSWIIC {								// - Template member functions
	void init(void);							// Init port
	T_IIC_ERROR start(void);					// Send start bit
	unsigned ex(unsigned);						// Exchange data
	T_IIC_ERROR stop(void); 					// Send stop bit
												// Write byte and verify - return incorrect bits
	inline unsigned write(unsigned n) { return ex(n) ^ n ^ 0x100; }
												// Read a byte and ACK
	inline unsigned read(void) { return ex(0x1FF); }
												// Read a byte and NAK
	inline unsigned read_nak(void) { return ex(0x0FF); }
};

template <
		const uint8_t scl,
		const uint8_t sda,
		volatile unsigned char &pin,
		volatile unsigned char &pout,
		volatile unsigned char &pdir
> void CSWIIC<scl, sda, pin, pout, pdir>::init(void)
{											 	// --- Init port
	pout |= scl;								// Clock is output
	pdir |= scl;								// Clock idles high
	pdir &= ~sda;								// Data is input
}

template <
		const uint8_t scl,
		const uint8_t sda,
		volatile unsigned char &pin,
		volatile unsigned char &pout,
		volatile unsigned char &pdir
> T_IIC_ERROR CSWIIC<scl, sda, pin, pout, pdir>::start(void)
{												// --- IIC Start - SDA high -> low while SCL high
	pout &= ~sda;								// Ensure SDA can be driven low
	if(!(pin & sda)) {							// If SDA is currently low then set high during clock low
		pout &= ~scl;							// SCL low
		pdir &= ~sda;							// SDA high
		if(!(pin & sda)) return ie_start_high;	// Return error if SDA did not go high
		pout |= scl;							// SCL high
	} 											//
	pdir |= sda;			 					// SDA low
	if(pin & sda) return ie_start_low;			// Return error if SDA did not go low
	return ie_ok;								// Return success
}

template <
		const uint8_t scl,
		const uint8_t sda,
		volatile unsigned char &pin,
		volatile unsigned char &pout,
		volatile unsigned char &pdir
> unsigned CSWIIC<scl, sda, pin, pout, pdir>::ex(unsigned tx)
{												// --- IIC data exchange
	unsigned m = 0x80;							// Begin with msb
	unsigned rx = 0;							// Clear rx data
	do {										//
		pout &= ~scl;							// Clock low
		(tx & m) ? (pdir &= ~sda) : (pdir |= sda); // Set data
		pout |= scl;							// Clock high
		if(pin & sda) rx |= m;					// Get rx bit
	} while(m >>= 1);							// Shift bitmask, check if done
	pout &= ~scl;								// Clock low
	(tx & 0x0100) ? (pdir |= sda) : (pdir &= ~sda); // Set ACK/NAK
	pout |= scl;								// Clock high
	if(!(pin & sda)) rx |= 0x0100;				// Get ACK/NAK
	return rx;									// Return rx data
}

template <
		const uint8_t scl,
		const uint8_t sda,
		volatile unsigned char &pin,
		volatile unsigned char &pout,
		volatile unsigned char &pdir
> T_IIC_ERROR CSWIIC<scl, sda, pin, pout, pdir>::stop(void)
{												// --- IIC Stop - SDA low -> high while SCL high
												// If SDA is currently high or being driven low by
												//  IIC slave, then set low during lock low
	if((pin & sda) || !(pdir & sda)) {			//
		pout &= ~scl;							// SCL low
		pdir |= sda;							// SDA low
		if(pin & sda) return ie_stop_low;		// Return error if SDA did no go low
		pout |= scl;							// SCL high
	}											//
	pdir &= ~sda;			 					// SDA high
	if(!(pin & sda)) return ie_stop_high;		// Return error if SDA did not go high
	return ie_ok;								// Return success
}
